from loaders import load_gz_csv_to_df
from forecasting_mle_decathlon.processing.processing_pipeline import (
    create_processing_pipeline
)
from forecasting_mle_decathlon.modeling.modeling_pipeline import (
    create_modeling_pipeline
)
from forecasting_mle_decathlon.modeling.utils import create_holidays_df
from sklearn.pipeline import Pipeline

import warnings
warnings.simplefilter("ignore")


if __name__ == '__main__':
    bu_df = load_gz_csv_to_df('data/bu_feat.csv.gz')
    train_df = load_gz_csv_to_df('data/train.csv.gz')
    test_df = load_gz_csv_to_df('data/train.csv.gz')

    holidays_df = create_holidays_df(2012, 2017)

    processing_pipeline = create_processing_pipeline(bu_df=bu_df)
    modeling_pipeline = create_modeling_pipeline(holidays_df=holidays_df)
    stages = processing_pipeline + modeling_pipeline
    pipe = Pipeline(stages)

    # FIXME: to remove it's just for test on two bu and dpt
    train_df_two_bu_dpt = train_df[
        (
            (train_df.but_num_business_unit == 64) |
            (train_df.but_num_business_unit == 119)
        ) &
        (train_df.dpt_num_department == 127)
        ]

    models = pipe.fit(train_df_two_bu_dpt)

    predictions = models.predict(X=train_df_two_bu_dpt)



from forecasting_mle_decathlon.processing.processing import (
    merge_two_df,
    remove_features,
    cast_series_to_datetime,
    create_year_and_week_column
)
import pandas as pd
from pandas.testing import assert_frame_equal
import pytest


@pytest.fixture
def df():
    return pd.DataFrame(
        [[1, "2021-09-10", 10], [2, "2021-10-10", 20]],
        columns=["id", "date", "value"]
    )


@pytest.fixture
def df_date(df):
    df["date"] = pd.to_datetime(df["date"])
    return df


def test_merge_two_df():
    # GIVEN
    df1 = pd.DataFrame(
        [[1, 1], [2, 1]],
        columns=["id", "value_df1"]
    )
    df2 = pd.DataFrame(
        [[1, 1], [2, 1]],
        columns=["id", "value_df2"]
    )
    expected = pd.DataFrame(
        [[1, 1, 1], [2, 1, 1]],
        columns=["id", "value_df1", "value_df2"]
    )
    merge_on = "id"

    # WHEN
    merge_result = merge_two_df(df1, df2, on=merge_on)

    # THEN
    assert_frame_equal(merge_result, expected)


def test_remove_features(df):
    # GIVEN
    feature_to_remove = ["date"]
    expected_columns = ["id", "value"]

    # WHEN
    result_df = remove_features(df, feature_to_remove)

    # THEN
    assert result_df.columns.values.tolist() == expected_columns


def test_cast_series_to_datetime(df):
    # GIVEN
    column_to_cast = "date"
    date_format = "%Y-%m-%d"
    expected_date_series = pd.Series(
        ["2021-09-10", "2021-10-10"]
    )
    expected_date_series = pd.to_datetime(expected_date_series)

    # WHEN
    result_series = cast_series_to_datetime(
        df,
        column_to_cast,
        date_format
    )[column_to_cast]

    # THEN
    assert (result_series.values.tolist() ==
            expected_date_series.values.tolist())


def test_create_year_and_week_column(df_date):
    # GIVEN
    column_name = "date"
    expected_df = pd.DataFrame(
        [[2021, 36], [2021, 40]],
        columns=["year", "week_number"]
    )
    # WHEN
    result_df = create_year_and_week_column(
        df_date,
        column_name
    )[["year", "week_number"]]

    # THEN
    assert_frame_equal(result_df, expected_df)

    pass



class DataframeFunctionTransformer():
    """
    Generic class to create custom estimator/transformer
    """
    def __init__(self, func, **kwargs):
        self.func = func
        self.parameters = kwargs

    def transform(self, input_df):
        return self.func(input_df, **self.parameters)

    def fit(self, X, y=None, **fit_params):
        return self



import pandas as pd
from typing import List, Tuple
from forecasting.processing.data_transformer import (
    DataframeFunctionTransformer
)
from forecasting.processing.processing import (
    merge_two_df,
    remove_features,
    cast_series_to_datetime,
    create_year_and_week_column,
    create_df_for_modeling
)


def create_processing_pipeline(
        bu_df: pd.DataFrame
) -> List[Tuple[str, DataframeFunctionTransformer]]:

    return [
        (
            'create_df',
            DataframeFunctionTransformer(
                merge_two_df,
                df2=bu_df,
                on="but_num_business_unit"
            )
        ),
        ('remove_correlated_features', DataframeFunctionTransformer(
            remove_features,
            features=["but_postcode", "zod_idr_zone_dgr"])
         ),
        (
            'cast_str_to_datetime',
            DataframeFunctionTransformer(
                cast_series_to_datetime,
                column_name="day_id"
            )
        ),
        ('create_year_and_week_number', DataframeFunctionTransformer(
            create_year_and_week_column,
            column_name="day_id")
         ),
        (
            'create_df_for_modeling',
            DataframeFunctionTransformer(create_df_for_modeling)
        ),
    ]



import pandas as pd
import numpy as np
from vacances_scolaires_france import SchoolHolidayDates


def create_holidays_df(start_year: int, end_year: int) -> pd.DataFrame:
    years = np.arange(start_year, end_year + 1)
    d = SchoolHolidayDates()
    holidays_list = []
    for holiday_year in years:
        holidays_list += list(d.holidays_for_year(holiday_year).keys())
    df = pd.DataFrame(holidays_list, columns=["ds"])
    df["holiday_name"] = "fr_holiday"
    df.rename(columns={"holiday_name": "holiday"}, inplace=True)
    return df



import pandas as pd


def load_gz_csv_to_df(filepath: str) -> pd.DataFrame:
    try:
        return pd.read_csv(
            filepath,
            compression='gzip',
            header=0, sep=',',
            error_bad_lines=False
        )
    except FileNotFoundError:
        exit(1)



import pandas as pd
from typing import List


def merge_two_df(
        df1: pd.DataFrame,
        df2: pd.DataFrame,
        on: str
) -> pd.DataFrame:
    return df1.merge(df2, on=on)


def remove_features(df: pd.DataFrame, features: List[str]) -> pd.DataFrame:
    return df.drop(columns=features, axis=1)


def cast_series_to_datetime(
        df: pd.DataFrame,
        column_name: str,
        date_format: str = "%Y-%m-%d"
) -> pd.DataFrame:
    df[column_name] = pd.to_datetime(df[column_name], format=date_format)
    return df


def create_year_and_week_column(
        df: pd.DataFrame,
        column_name: str
) -> pd.DataFrame:
    df["year"] = df[column_name].dt.year
    df["week_number"] = df[column_name].dt.week
    return df


def sum_turnover_by_week_bu_dpt(
        df: pd.DataFrame,
        features: List[str] =
        ["but_num_business_unit", "dpt_num_department", "year", "week_number"],
        column_to_sum: str = "turnover"
) -> pd.DataFrame:
    return (
        df
        .groupby(features, as_index=False)
        .agg({column_to_sum: "sum"})
    )


def create_id_bu_dpt(df: pd.DataFrame) -> pd.DataFrame:
    df["but_bu_dpt_id"] = (
            df["but_num_business_unit"].astype(str)
            + "-"
            + df["dpt_num_department"].astype(str)
    )
    return df


def create_id_year_week_number(df: pd.DataFrame) -> pd.DataFrame:
    df["ds"] = (
            df["year"].astype(str)
            + "-"
            + df["week_number"].astype(str)
    )
    return df


def create_date_from_year_and_week_number(df: pd.DataFrame) -> pd.DataFrame:
    df["ds"] = (
        pd.to_datetime(
            df["ds"] + "-1",
            format="%Y-%W-%w"
        )
    )
    return df


def rename_taget_column(
        df: pd.DataFrame,
        target_col_to_rename: str
) -> pd.DataFrame:
    return df.rename(columns={target_col_to_rename: "y"})


def create_store_dpt_df(df: pd.DataFrame) -> List[pd.DataFrame]:
    return [df[df["but_bu_dpt_id"] == i] for i in df["but_bu_dpt_id"].unique()]


def select_columns_to_fit(
        dfs: List[pd.DataFrame],
        columns_to_fit: List[str] =
        ["ds", "y", "but_bu_dpt_id"]
) -> List[pd.DataFrame]:
    return [df[columns_to_fit] for df in dfs]


def remove_n_last_values(
        dfs: List[pd.DataFrame],
        n_value_to_remove: int = 8
) -> List[pd.DataFrame]:
    return [df.drop(df.tail(n_value_to_remove).index) for df in dfs]


def create_df_for_modeling(df: pd.DataFrame) -> List[pd.DataFrame]:
    df = sum_turnover_by_week_bu_dpt(df)
    df = create_id_bu_dpt(df)
    df = create_id_year_week_number(df)
    df = create_date_from_year_and_week_number(df)
    df = rename_taget_column(df, target_col_to_rename="turnover")

    df_list = create_store_dpt_df(df)
    df_list = remove_n_last_values(df_list)
    df_list = select_columns_to_fit(df_list)

    return df_list



import pandas as pd
from forecasting.modeling.prophet_estimator import ProphetEstimator
from typing import List, Tuple


def create_modeling_pipeline(
        holidays_df: pd.DataFrame
) -> List[Tuple[str, ProphetEstimator]]:
    return [
        ('model', ProphetEstimator(holidays_df))
    ]



from sklearn.base import BaseEstimator
from prophet import Prophet


class ProphetEstimator(BaseEstimator):
    def __init__(self, holidays_df):
        super().__init__()
        self.holidays_df = holidays_df
        self.models = {}

    def fit(self, dfs, Y=None):
        for df in dfs:
            df = df.reset_index(drop=True)
            model = Prophet(
                holidays=self.holidays_df,
                holidays_prior_scale=1
            )
            model.fit(df[["ds", "y"]])
            key = df.loc[0]["but_bu_dpt_id"]
            self.models[key] = model
        return self

    def predict(self, X=None):
        predictions = {}
        for key in self.models.keys():
            future_df = (
                self.models
                    .get(key)
                    .make_future_dataframe(periods=8, freq='W')
            )
            predictions[key] = self.models.get(key).predict(future_df)
        return predictions



from setuptools import setup

with open('requirements.txt') as f:
    required_packages = f.read().splitlines()

setup(
   name='forecasting-mle-decathlon',
   version='0.1.1',
   author='Rugery Pierrick',
   author_email='rugery.pierrick@fake.com',
   packages=['forecasting_mle_decathlon'],
   description='Package to forecast sales',
   install_requires=required_packages,
)
